import numpy as np
from ctypes import *

TSTRING = c_int(16)
TLOGICAL   = c_int( 14)
TBYTE      = c_int( 11)
TSTRING    = c_int( 16)
TUSHORT    = c_int( 20)
TSHORT     = c_int( 21)
TUINT      = c_int( 30)
TINT       = c_int( 31)
TULONG     = c_int( 40)
TLONG      = c_int( 41)
TFLOAT     = c_int( 42)
TLONGLONG  = c_int( 81)
TDOUBLE    = c_int( 82)
TCOMPLEX   = c_int( 83)
TDBLCOMPLEX= c_int(163)


BINARY_TBL = c_int(2)


TFORM_NP = {'L': 'i1', 'B': 'u1', 'I': 'i2', 'J': 'i4', 'K': 'i8', 'E': 'f4',
              'D': 'f8', 'C': 'c8', 'M': 'c16', 'A': 'a'}

NP_TFORM = dict([[v,k] for k,v in TFORM_NP.items()])

TFORM_FITS = {
    'L':TLOGICAL,
    'B':TBYTE,
    'I':TSHORT,
    'J':TLONG,
    'K':TLONGLONG,
    'E':TFLOAT,
    'D':TDOUBLE,
    'C':TCOMPLEX,
    'M':TDBLCOMPLEX,
    'A':TSTRING
}

TFORM_CTYPES = {
    'i1':c_bool, 
    'u1':c_uint8, 
    'i2':c_int16, 
    'i4':c_int32, 
    'i8':c_int64, 
    'f4':c_float, 
    'f8':c_double, 
    'a':c_char,
    'c8':None,
    'c16':None
}



class cache(object):    
    '''Computes attribute value and caches it in the instance.
    Python Cookbook (Denis Otkidach) http://stackoverflow.com/users/168352/denis-otkidach
    This decorator allows you to create a property which can be computed once and
    accessed many times. Sort of like memoization.

    '''
    def __init__(self, method, name=None):
        # record the unbound-method and the name
        self.method = method
        self.name = name or method.__name__
        self.__doc__ = method.__doc__
    def __get__(self, inst, cls):
        # self: <__main__.cache object at 0xb781340c>
        # inst: <__main__.Foo object at 0xb781348c>
        # cls: <class '__main__.Foo'>       
        if inst is None:
            # instance attribute accessed on class, return self
            # You get here if you write `Foo.bar`
            return self
        # compute, cache and return the instance's attribute value
        result = self.method(inst)
        # setattr redefines the instance's attribute so this doesn't get called again
        setattr(inst, self.name, result)
        return result



from pycfitsio import *
import numpy as np
import os
import unittest

hdu_content = np.empty(1000, dtype=np.dtype([
                                    ('signal', np.double),
                                    ('flag', np.uint8)
                                    ]))
hdu_content['signal'] = np.arange(1000)
hdu_content['flag'] = np.ones(1000)

class TestPyCfitsIoRead(unittest.TestCase):

    def setUp(self):
        self.filename = "../debug/data.fits"
        self.hdus = 1
        self.hduname = "DATA"
        self.len = 1000

    def test_open_file(self):
        f = open(self.filename)
        self.assertEqual(self.filename, f.filename)
        self.assertTrue(isinstance(f, File))

    def test_read_hdus(self):
        f = open(self.filename)
        hdus = f.HDUs
        self.assertEqual(len(hdus), 1)
        self.assertEqual(hdus.values()[0].name, self.hduname)
        self.assertEqual(f[0].name, self.hduname)
        self.assertEqual(hdus[self.hduname].name, self.hduname)

    def test_move(self):
        f = open(self.filename)
        f.move(f[0].name)
        self.assertEqual(f.current_HDU, self.hduname)

    def test_colnames(self):
        f = open(self.filename)
        self.assertEqual(f[0].column_names, hdu_content.dtype.names)


    def test_read_col(self):
        f = open(self.filename)
        h = f[0]
        for i, name in enumerate(hdu_content.dtype.names):
            np.testing.assert_array_almost_equal(h.read_column(i), hdu_content[name])
            np.testing.assert_array_almost_equal(h.read_column(name), hdu_content[name])

    def test_all_cols(self):
        f = open(self.filename)
        data = f[0].read_all()
        self.assertTrue(isinstance(data, np.ndarray))
        self.assertEqual(data.dtype.names, hdu_content.dtype.names)
        for name in hdu_content.dtype.names:
            np.testing.assert_array_almost_equal(data[name], hdu_content[name])

    def test_read(self):
        data, header = read(self.filename)
        self.assertTrue(isinstance(data, np.ndarray))
        self.assertEqual(data.dtype.names, hdu_content.dtype.names)
        for name in hdu_content.dtype.names:
            np.testing.assert_array_almost_equal(data[name], hdu_content[name])

class TestPyCfitsIoWrite(unittest.TestCase):

    def setUp(self):
        self.filename = "../debug/newdata.fits"
        try:
            os.remove(self.filename)
        except:
            pass

    def test_create(self):
        f = create(self.filename)
        f.write_HDU("NEWDATA", hdu_content)
        f.close()
        self.assertTrue(isinstance(f, File))
        self.assertTrue(os.path.exists(self.filename))

    def test_write(self):
        fw = create(self.filename)
        fw.write_HDU("NEWDATA", hdu_content)
        fw.close()
        f = open(self.filename)
        data = f[0].read_all()
        self.assertTrue(isinstance(data, np.ndarray))
        self.assertEqual(f[0].name, "NEWDATA")
        for name in data.dtype.names:
            np.testing.assert_array_almost_equal(data[name], hdu_content[name])

    def test_repeat_tform(self):
        array = open('../debug/tform.fits')[0].read_column('data')
        np.testing.assert_array_almost_equal(array, np.arange(10000))

    def tearDown(self):
        try:
            os.remove(self.filename)
        except:
            pass

if __name__ == '__main__':
    unittest.main()



import numpy as np
import exceptions
try:
    from collections import OrderedDict
except exceptions.ImportError:
    from ordereddict import OrderedDict
from contextlib import closing
from ctypes import *

from decorator import cache

from constants import *
import ctypes.util

_cfitsio = CDLL(ctypes.util.find_library("cfitsio"))
NULL = c_double(0.)

class CfitsioError(exceptions.Exception):
#TODO verbose error messages
    pass

def open(filename):
    """Opens fits file and returns a File object"""
    f = File(str(filename))
    f.open()
    return f

def read(filename, HDU=0):
    """Quick function for reading one HDU and header

    HDU is name or index of HDU, default 0
    Returns the data array and a header dictionary"""

    #TODO context manager
    with closing(open(filename)) as f:
        return f[HDU].read_all(), f[HDU].header

def create(filename):
    """Create a new fits file and returns a File object"""
    f = File(filename)
    f.create()
    return f

def check_status(status):
    if status.value != 0:
        raise CfitsioError("CFITSIO error with code %d" % status.value)

def run_check_status(function, *args):
    status = c_int()
    args = list(args)
    args.append(byref(status))
    function(*args)
    check_status(status)

class File(object):

    def __init__(self, filename):
        self.ptr = c_voidp()
        self.filename = filename
        self.current_HDU = None

    def __repr__(self):
        return "Fits FILE: %s" % self.filename

    def open(self):
        run_check_status(_cfitsio.ffopen, byref(self.ptr), self.filename, False)

    def create(self):
        run_check_status(_cfitsio.ffinit, byref(self.ptr), '!'+self.filename)

    def close(self):
        run_check_status(_cfitsio.ffclos, self.ptr)

    def __getitem__(self, key):
        """Returns HDU by name or index"""
        if isinstance(key, int):
            return self.HDUs.values()[key]
        else:
            return self.HDUs[key]

    @cache
    def HDUs(self):
        _HDUs = OrderedDict()
        hdunum = c_int()
        run_check_status(_cfitsio.ffthdu, self.ptr, byref(hdunum))
        hdutype = c_int()
        for i in range(2, hdunum.value+1):
            run_check_status(_cfitsio.ffmahd, self.ptr, c_int(i), byref(hdutype))
            hdu = HDU(i-2, file=self)
            _HDUs[hdu.name] = hdu
        return _HDUs

    def move(self, name):
        if self.current_HDU != name:
            if isinstance(name, int):
                hdutype = c_int(0)
                run_check_status(_cfitsio.ffmahd, self.ptr, c_int(name+2), byref(hdutype))
            elif name.startswith('HDU'):
                self.move(int(name[-1]))
            else:
                run_check_status(_cfitsio.ffmnhd, self.ptr, BINARY_TBL, name, False)
            self.current_HDU = name


    def write_HDU(self, name, data):
        """Data must be a numpy array with named dtype"""
        column_names = data.dtype.names
        keywords_t = c_char_p * len(column_names)
        ttype = keywords_t(*map(c_char_p, column_names))
        tform = keywords_t(*[NP_TFORM[data[colname].dtype.str[1:]] for colname in column_names])
        run_check_status(_cfitsio.ffcrtb, self.ptr, BINARY_TBL, c_longlong(0), c_int(len(column_names)), byref(ttype), byref(tform), byref(NULL), c_char_p(name))

        buffer_size = c_long(1)
        run_check_status(_cfitsio.ffgrsz, self.ptr, byref(buffer_size))

        for k in range(0,data.size,buffer_size.value):
            if (k+buffer_size.value) > data.size:
                buffer_size = c_long(data.size - k)
            for i, colname in enumerate(column_names):
                np_dtype = data[colname].dtype.str[1:]
                coltform = NP_TFORM[np_dtype]
                contiguous_data = np.ascontiguousarray(data[colname][k:])
                run_check_status(_cfitsio.ffpcl, self.ptr, TFORM_FITS[coltform], c_int(i+1), c_longlong(1+k), c_longlong(1), c_longlong(buffer_size.value), contiguous_data.ctypes.data_as(POINTER(TFORM_CTYPES[np_dtype])))

    def write_HDU_dict(self, name, data):
        """Data must be an OrderedDict of arrays"""
        keywords_t = c_char_p * len(data)
        ttype = keywords_t(*map(c_char_p, data.keys()))
        data_length = len(data.values()[0])
        tform = keywords_t(*[NP_TFORM[col.dtype.str[1:]] for col in data.values()])
        run_check_status(_cfitsio.ffcrtb, self.ptr, BINARY_TBL, c_longlong(0), c_int(len(data)), byref(ttype), byref(tform), byref(NULL), c_char_p(name))
    
        buffer_size = c_long(1)
        run_check_status(_cfitsio.ffgrsz, self.ptr, byref(buffer_size))
    
        for k in range(0,len(data.values()[0]),buffer_size.value):
            if (k+buffer_size.value) > data_length:
                buffer_size = c_long(data_length - k)
            for i, (colname, colarray) in enumerate(data.iteritems()):
                np_dtype = colarray.dtype.str[1:]
                coltform = NP_TFORM[np_dtype]
                run_check_status(_cfitsio.ffpcl, self.ptr, TFORM_FITS[coltform], c_int(i+1), c_longlong(1+k), c_longlong(1), c_longlong(buffer_size.value), colarray[k:].ctypes.data_as(POINTER(TFORM_CTYPES[np_dtype])))

class HDU(object):

    def __init__(self, num=0, file=None):
        self.num = num
        self.file = file
        try:
            self.name = self.get_header_keyword("EXTNAME").strip()
        except CfitsioError:
            self.name = "HDU%d" % num

    #TODO implement self.file.move as a decorator
    #def move(self):
    #    def newf():
    #        self.file.move(self.name)

    def __repr__(self):
        return "HDU: %s" % self.name

    def get_header_key(self, name):
        self.file.move(self.num)
        value = c_long()
        run_check_status(_cfitsio.ffgky, self.file.ptr, TLONG, name, byref(value), False)
        return value.value

    def get_header_keyword(self, name):
        self.file.move(self.num)
        value = (c_char*50)()
        run_check_status(_cfitsio.ffgky, self.file.ptr, TSTRING, name, byref(value) , False)
        return value.value.strip()

    def read_column(self, num):
        self.file.move(self.num)
        if isinstance(num, str):
            num = self.column_names.index(num)
        array = np.empty(self.length, dtype=self.dtype[num])
        run_check_status(_cfitsio.ffgcv, self.file.ptr, TFORM_FITS[self.fits_datatypes[num]], c_int(num+1), c_longlong(1), c_longlong(1), c_longlong(self.length), byref(NULL), array.ctypes.data_as(POINTER(TFORM_CTYPES[self.dtype[num].str[1:]])), byref(NULL))
        return array

    @property
    def column_names(self):
        return self.dtype.names

    @cache
    def length(self):
        self.file.move(self.name)
        length = self.get_header_key("NAXIS2")
        #check repeat
        tform = self.get_header_keyword("TFORM1")
        if len(tform) > 1:
            length *= long(tform[:-1])
        return length

    @cache
    def fits_datatypes(self):
        self.file.move(self.num)
        return [self.get_header_keyword("TFORM%d" % (num+1))[-1] for num in range(self.num_columns)]

    @cache
    def num_columns(self):
        self.file.move(self.num)
        return self.get_header_key("TFIELDS")

    @cache
    def header(self):
        self.file.move(self.num)
        keysexist, numkeys = c_int(0), c_int(0)
        run_check_status(_cfitsio.ffghsp, self.file.ptr, byref(keysexist), byref(numkeys))
        header = {}
        for i in range(1, keysexist.value+1):
            name = (c_char*50)()
            value = (c_char*50)()
            comment = (c_char*50)()
            run_check_status(_cfitsio.ffgkyn, self.file.ptr, c_int(i), byref(name), byref(value), byref(comment))
            header[name.value] = value.value.replace("'","").strip()
        return header

    @cache
    def dtype(self):
        self.file.move(self.num)
        data_dtype = []
        for num, fits_datatype in enumerate(self.fits_datatypes):
            data_dtype.append((self.get_header_keyword("TTYPE%d" % (num+1)), TFORM_NP[fits_datatype]))
        return np.dtype(data_dtype)

    def read_all(self, asodict=False):
        """Read columns into numpy array"""
        if asodict:
            data = OrderedDict()
        else:
            data = np.empty(self.length, dtype=self.dtype)
        #TODO implement buffered read
        for i, name in enumerate(self.column_names):
            data[name] = self.read_column(i)
        return data

if __name__ == '__main__':
    f = open("../test/data.fits")
    print(f.HDUs)
    print(f['DATA'])
    self=f[0]
    data = f["DATA"].read_column('signal')
    a = f[0].read_all()
    self = create('../test/newdata.fits')
    self.write_HDU('newdata',a)
    self.close()
    print(f[0].header)
    #all_data = f["DATA"].read_all()



from distutils.core import setup

setup(
    name = "pycfitsio",
    version = "0.1.0",

    packages=['pycfitsio', 'pycfitsio.test'],

    author = "Andrea Zonca",
    author_email = "code@andreazonca.com",
    description = "Simple FITS files I/O package",
    license='LICENSE.txt',
    keywords = "fits, astrophysics, pyfits",
    url = "http://github.com/zonca/pycfitsio",
    classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: GNU General Public License (GPL)',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: POSIX',
          'Programming Language :: Python',
          'Topic :: Scientific/Engineering :: Physics',
          ],
    long_description=open('README').read(),

)



#log# Automatic Logger file. *** THIS MUST BE THE FIRST LINE ***
#log# DO NOT CHANGE THIS LINE OR THE TWO BELOW
#log# opts = Struct({'__allownew': True, 'logfile': 'ipython_log.py', 'pylab': 1})
#log# args = []
#log# It is safe to make manual edits below here.
#log#-----------------------------------------------------------------------
import os
#?os.remove
